/*
 * IRQ-handling code for TiVo devices
 *
 * Copyright (C) 1992 Linus Torvalds
 * Copyright (C) 1994 - 2000 Ralf Baechle
 * Copyright (C) 2000 TiVo Inc. All Rights Reserved.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/plog.h>
#include <linux/module.h>
#include <linux/sw_watchdog.h>
#include <linux/fastbufs.h>

#include <asm/mipsregs.h>
#include <asm/ptrace.h>
#include <asm/bcm7020.h>
#include <asm/px_hw.h>
#include <asm/atomic.h>

extern int fastbuf_initialized;
extern int tivo_monotonic_initialized;
extern int systime_initialized;

extern asmlinkage void tivo_handle_irq(void);
extern void do_fpgairq(struct pt_regs *regs);
 
atomic_t irq_err_count;

static unsigned int bcm_pir_cache, bcm_ie_kernel, ms_im_cache, ms_him_cache, ms_tim_cache, ms_gim_cache;
static struct irqaction *irq_action[BCM_NR_IRQS];
static int irq_depth[BCM_NR_IRQS];

#ifdef CONFIG_TIVO_DEVEL

int debug_irqs;

static int __init dbgirq_setup(char *str)
{
	get_option(&str, &debug_irqs);
	return 1;
}

__setup("dbgirq=", dbgirq_setup);

#define DEBUG_IRQS debug_irqs

#else
#define DEBUG_IRQS 0
#endif

static inline void mask_irq(unsigned int irq)
{
	unsigned int status;

	if (irq < 8) {
		status = read_32bit_cp0_register(CP0_STATUS);
		status &= ~(1 << (irq + 8));
		write_32bit_cp0_register(CP0_STATUS, status);
#ifdef CONFIG_SW_WATCHDOG
		_gbl_sti_cli_flags &= ~(1 << (irq + 8));
#endif
	} else if (irq  < 40) {
		bcm_ie_kernel &= ~(1 << (irq - 8));
		*BCM_INTR_ENA &= ~(1 << (irq - 8));
	} else if (irq < MEDIASWITCH_IBASE) {
		bcm_pir_cache &= ~(1 << (irq - 40));
		*BCM_PCI_INTR_REG = bcm_pir_cache;
	} else if (irq < MEDIASWITCH_IHBASE) {
		ms_im_cache &= ~(1 << (irq - MEDIASWITCH_IBASE));
		MsWrite(MS_IM_REG, ms_im_cache);
	} else if (irq < MEDIASWITCH_ITBASE) {
		ms_him_cache &= ~(1 << (irq - MEDIASWITCH_IHBASE));
		MsWrite(MS_HEIM_REG, ms_him_cache);
	} else if (irq < MEDIASWITCH_IGBASE) {
		ms_tim_cache &= ~(1 << (16 + irq - MEDIASWITCH_ITBASE));
		MsWrite(MS_TPIM_REG, ms_tim_cache);
	} else {
		ms_gim_cache &= ~(1 << (irq - MEDIASWITCH_IGBASE));
		MsWrite(MS_GPIOIM_REG, ms_gim_cache);
	}
	barrier();
}

static inline void mask_and_ack_irq(unsigned int irq)
{
	unsigned int status;

	if (irq < 8) {
		status = read_32bit_cp0_register(CP0_STATUS);
		status &= ~(1 << (irq + 8));
		write_32bit_cp0_register(CP0_STATUS, status);
#ifdef CONFIG_SW_WATCHDOG
		_gbl_sti_cli_flags &= ~(1 << (irq + 8));
#endif
	} else if (irq  < 40) {
		bcm_ie_kernel &= ~(1 << (irq - 8));
		*BCM_INTR_ENA &= ~(1 << (irq - 8));
		barrier();
		*BCM_INTR_CLEAR = 1 << (irq - 8);
	} else if (irq < MEDIASWITCH_IBASE) {
		bcm_pir_cache &= ~(1 << (irq - 40));
		*BCM_PCI_INTR_REG = bcm_pir_cache;
		*BCM_PCI_INTR_STAT = 1 << (irq - 40);
	} else if (irq < MEDIASWITCH_IHBASE) {
		ms_im_cache &= ~(1 << (irq - MEDIASWITCH_IBASE));
		MsWrite(MS_IM_REG, ms_im_cache);
		MsWrite(MS_ISR_REG, 1 << (irq - MEDIASWITCH_IBASE));
	} else if (irq < MEDIASWITCH_ITBASE) {
		ms_him_cache &= ~(1 << (irq - MEDIASWITCH_IHBASE));
		MsWrite(MS_HEIM_REG, ms_him_cache);
		MsWrite(MS_HEIS_REG, 1 << (irq - MEDIASWITCH_IHBASE));
	} else if (irq < MEDIASWITCH_IGBASE) {
		ms_tim_cache &= ~(1 << (16 + irq - MEDIASWITCH_ITBASE));
		MsWrite(MS_TPIM_REG, ms_tim_cache);
		MsWrite(MS_TPIS_REG, 1 << (16 + irq - MEDIASWITCH_ITBASE));
	} else {
		ms_gim_cache &= ~(1 << (irq - MEDIASWITCH_IGBASE));
		MsWrite(MS_GPIOIM_REG, ms_gim_cache);
		MsWrite(MS_GPIOIS_REG, 1 << (irq - MEDIASWITCH_IGBASE));
	}
	barrier();
}

static inline void ack_irq(unsigned int irq)
{
	if (irq >= 8) {
		if (irq  < 40) {
			*BCM_INTR_CLEAR = 1 << (irq - 8);
		} else if (irq < MEDIASWITCH_IBASE) {
			*BCM_PCI_INTR_STAT = 1 << (irq - 40);
		} else if (irq < MEDIASWITCH_IHBASE) {
			MsWrite(MS_ISR_REG, 1 << (irq - MEDIASWITCH_IBASE));
		} else if (irq < MEDIASWITCH_ITBASE) {
			MsWrite(MS_HEIS_REG, 1 << (irq - MEDIASWITCH_IHBASE));
		} else if (irq < MEDIASWITCH_IGBASE) {
			MsWrite(MS_TPIS_REG, 1 << (16 + irq - MEDIASWITCH_ITBASE));
		} else {
			MsWrite(MS_GPIOIS_REG, 1 << (irq - MEDIASWITCH_IGBASE));
		}
	}
	barrier();
}

static inline void unmask_irq(unsigned int irq)
{
	unsigned int status;

	if (irq < 8) {
		status = read_32bit_cp0_register(CP0_STATUS);
		status |= 1 << (irq + 8);
		write_32bit_cp0_register(CP0_STATUS, status);
#ifdef CONFIG_SW_WATCHDOG
		_gbl_sti_cli_flags |= 1 << (irq + 8);
#endif
	} else if (irq  < 40) {
		bcm_ie_kernel |= 1 << (irq - 8);
		*BCM_INTR_ENA |= 1 << (irq - 8);
	} else if (irq < MEDIASWITCH_IBASE) {
		bcm_pir_cache |= 1 << (irq - 40);
		*BCM_PCI_INTR_REG = bcm_pir_cache;
	} else if (irq < MEDIASWITCH_IHBASE) {
		ms_im_cache |= 1 << (irq - MEDIASWITCH_IBASE);
		MsWrite(MS_IM_REG, ms_im_cache);
	} else if (irq < MEDIASWITCH_ITBASE) {
		ms_him_cache |= 1 << (irq - MEDIASWITCH_IHBASE);
		MsWrite(MS_HEIM_REG, ms_him_cache);
	} else if (irq < MEDIASWITCH_IGBASE) {
		ms_tim_cache |= 1 << (16 + irq - MEDIASWITCH_ITBASE);
		MsWrite(MS_TPIM_REG, ms_tim_cache);
	} else {
		ms_gim_cache |= 1 << (irq - MEDIASWITCH_IGBASE);
		MsWrite(MS_GPIOIM_REG, ms_gim_cache);
	}
	barrier();
}

void disable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	if (DEBUG_IRQS >= 3) {
		printk("dis<%d>", irq_nr);
	}
	if (irq_nr >= BCM_NR_IRQS)
		return;
	save_and_cli(flags);
	if (irq_depth[irq_nr]++ < 0)
		mask_irq(irq_nr);
	restore_flags(flags);
}

void enable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	if (DEBUG_IRQS >= 3) {
		printk("ena<%d>", irq_nr);
	}
	if (irq_nr >= BCM_NR_IRQS)
		return;
	save_and_cli(flags);
	if (--irq_depth[irq_nr] < 0)
		unmask_irq(irq_nr);
	restore_flags(flags);
}

EXPORT_SYMBOL(disable_irq);
EXPORT_SYMBOL(enable_irq);

int get_irq_list(char *buf)
{
	int i, len = 0;
	struct irqaction *action;

	for (i = 0 ; i < BCM_NR_IRQS ; i++) {
		action = irq_action[i];
		if (!action)
			continue;
		len += sprintf(buf + len, "%2d: %8d %c %s",
			i, kstat.irqs[0][i],
			(action->flags & SA_INTERRUPT) ? '+' : ' ',
			action->name);
		for (action = action->next; action; action = action->next) {
			len += sprintf(buf + len, ",%s %s",
				(action->flags & SA_INTERRUPT) ? " +" : "",
				action->name);
		}
		len += sprintf(buf + len, "\n");
	}
	return len;
}

/*
 * do_IRQ is called by the low-level interrupt handling code with the
 * appropriate IRQ number
 */

asmlinkage void do_IRQ(int irq, struct pt_regs *regs)
{
	struct irqaction *action;
	int cpu;
#ifdef CONFIG_SW_WATCHDOG
	/* This is a bug fix for the following condition:
	 * You're in syscall context or BH context, turned off
	 * interrupts, and a wrap int fires.  It's very broken
	 * to attempt to run soft irq's in that case.  Since a
	 * wrap int is the only case where this can occur, we
	 * just turn off dealing with softirq's in the case that
	 * a wrap fired.
	 */
	int disableSoft=0;
	if (irq == CPU_IRQ_INT5)
		disableSoft=1;
#endif
	mask_and_ack_irq(irq);
	cpu = smp_processor_id();
	irq_enter(cpu, irq);
#ifdef CONFIG_SW_WATCHDOG
	sw_watchdog_irq_enter(cpu, irq);
#endif
	kstat.irqs[cpu][irq]++;

	action = irq_action[irq];

	if (DEBUG_IRQS >= 2) {
		if (irq == BCM_IRQ_TMR_COMP && DEBUG_IRQS < 5) {
			static unsigned int count1;
			if (!(count1++ & 0xff)) {
				printk("*");
			}
		} else if (irq == CPU_IRQ_INT2) {
			if (DEBUG_IRQS >= 5) {
				printk("x%x", *BCM_INTR_STATUS & *BCM_INTR_ENA);
			} else {
				static unsigned int count2;
				if (!(count2++ & 0xff)) {
					printk("x%x", *BCM_INTR_STATUS & *BCM_INTR_ENA);
				}
			}
		} else {
			printk("[%d", irq);
		}
	}

	if (action != 0) {
		unsigned long flags = 0;

		plog_intr_enter(irq, regs->cp0_epc);
		do {
#ifdef CONFIG_SW_WATCHDOG
			/* WD Masks out everything but wrap count */
			__cli();
			/* WD Turns back on interupt enable bit: */
			__really_sti();
#endif
			if (!(action->flags & SA_INTERRUPT))
				__sti();
			action->handler(irq, action->dev_id, regs);
			__cli();
			flags |= action->flags;
			action = action->next;
		} while (action);
		if (flags & SA_SAMPLE_RANDOM)
			add_interrupt_randomness(irq);
		// if IRQ was disabled from within its own handler(s),
		// leave it masked, otherwise re-enable
		if (irq_depth[irq] < 0)
			unmask_irq(irq);
		plog_intr_leave(irq);
	} else if (DEBUG_IRQS >= 2) {	// report as unhandled
		printk("!");
	}
	if (DEBUG_IRQS >= 2) {
		if (irq != BCM_IRQ_TMR_COMP && irq != CPU_IRQ_INT2) {
			printk("]");
		} else if (DEBUG_IRQS >= 5) {
			if (irq == BCM_IRQ_TMR_COMP) {
				printk("]");
			} else if (irq == CPU_IRQ_INT2) {
				printk("X");
			}
		}
	}
    
    if ( fastbuf_initialized && fastbuf ){
        if ( tivo_monotonic_initialized ){
            tivo_monotonic_get_current(&fastbuf->t);
            fastbuf->rseq++;
        }
        if ( systime_initialized ){
            do_gettimeofday(&fastbuf->tv);
            fastbuf->rseq++;
        }
    }

	irq_exit(cpu, irq);

#ifdef CONFIG_SW_WATCHDOG
	if (!disableSoft) {
#endif
		if (softirq_pending(cpu))
			do_softirq();
#ifdef CONFIG_SW_WATCHDOG
	}
	/* Make CP0_STATUS sane again */
	{
		unsigned int cp0status = read_32bit_cp0_register(CP0_STATUS);
		cp0status &= ~(1); /* Clear interrupt enabled */
		cp0status |= (_gbl_sti_cli_flags & 0x7f00);
		write_32bit_cp0_register(CP0_STATUS, cp0status);
	}
#endif
}

int tivo_setup_irq(int irq, struct irqaction *new)
{
	int shared = 0;
	struct irqaction *old, **p;
	unsigned long flags;

	p = irq_action + irq;
	if ((old = *p) != NULL) {
		// Shared ints must all agree to share
		if (!(old->flags & new->flags & SA_SHIRQ))
			return -EBUSY;
		do {
			p = &old->next;
			old = *p;
		} while (old);
		shared = 1;
	}
	if (new->flags & SA_SAMPLE_RANDOM)
		rand_initialize_irq(irq);

	save_and_cli(flags);
	*p = new;

	if (!shared) {
		irq_depth[irq] = -1;
		ack_irq(irq);
		unmask_irq(irq);
	}
	restore_flags(flags);
	return 0;
}

int request_irq(unsigned int irq,
	void (*handler)(int, void *, struct pt_regs *),
	unsigned long irqflags,
	const char *devname,
	void *dev_id)
{
	int retval;
	struct irqaction *action;

	if (DEBUG_IRQS >= 1) {
		printk("request irq %d\n", irq);
	}
	if (irq >= BCM_NR_IRQS)
		return -EINVAL;
	if (!handler)
		return -EINVAL;
	// Only allow a MS IRQ if it's actually present...
	if (irq >= MEDIASWITCH_IBASE && !MsGetBase())
		return -EINVAL;

	action = (struct irqaction *)kmalloc(sizeof(struct irqaction), GFP_KERNEL);
	if (!action)
		return -ENOMEM;

	action->handler = handler;
	action->flags = irqflags;
	action->mask = 0;
	action->name = devname;
	action->next = NULL;
	action->dev_id = dev_id;

	retval = tivo_setup_irq(irq, action);

	if (retval)
		kfree(action);
	return retval;
}

void free_irq(unsigned int irq, void *dev_id)
{
	struct irqaction *action, **p;
	unsigned long flags;

	if (DEBUG_IRQS >= 1) {
		printk("free irq %d\n", irq);
	}
	if (irq >= BCM_NR_IRQS) {
		printk(KERN_ERR "Error - trying to free invalid IRQ %d\n", irq);
		return;
	}
	for (p = irq + irq_action; (action = *p) != NULL; p = &action->next) {
		if (action->dev_id != dev_id)
			continue;

		// Found it - now free it
		save_and_cli(flags);
		*p = action->next;
		if (!irq[irq_action])
			mask_irq(irq);
		restore_flags(flags);
		kfree(action);
		return;
	}
	printk(KERN_ERR "Trying to free free IRQ%d\n", irq);
}

unsigned long probe_irq_on(void)
{
	// not implemented
	return 0;
}

unsigned int probe_irq_mask(unsigned long val)
{
	// not implemented
	return 0;
}

int probe_irq_off(unsigned long irqs)
{
	// not implemented
	return 0;
}

int (*irq_cannonicalize)(int irq);

static int bcm_irq_cannonicalize(int irq)
{
	return irq;
}

static void bcm_illcyc_int(int irq, void *dev_id, struct pt_regs *regs)
{
#if 0 // MFK: not yet, bug in bcm7020 driver?
	die("CPU illegal cycle detected", regs);
#else
	static unsigned long last_err = 0;

	// don't flood console when we do a bunch of illegal writes
	if (last_err > jiffies - 10) {
		return;
	}
	last_err = jiffies;
	printk("CPU illegal cycle detected at EPC=0x%08lx\n", regs->cp0_epc);
#endif
}

static void dummy_int(int irq, void *dev_id, struct pt_regs *regs)
{
	return;
}

static struct irqaction cascade_shared =
	 { dummy_int, SA_INTERRUPT | SA_SHIRQ, 0, "cascade", dummy_int, NULL };

struct irqaction cascade =
	 { NULL, SA_INTERRUPT, 0, "cascade", NULL, NULL };

static struct irqaction bcm_illcyc_irq =
	 { bcm_illcyc_int, SA_INTERRUPT, 0, "illegal cycle", NULL, NULL };

void __init init_IRQ(void)
{
	unsigned int status;

	irq_cannonicalize = bcm_irq_cannonicalize;
	set_except_vector(0, tivo_handle_irq);

	status = read_32bit_cp0_register(CP0_STATUS);
	status &= 0xffff00ff;
	write_32bit_cp0_register(CP0_STATUS, status);

	// Ack and disable all PCI ints
	*BCM_PCI_INTR_REG = 0x0000;
	*BCM_PCI_INTR_STAT = 0xffff;
	barrier();
	// Ack and disable all BCM7020 ints
	*BCM_INTR_ENA = 0x00000000;
	barrier();
	*BCM_INTR_CLEAR = 0xffffffff;
	barrier();
	// Turn on int output to CPU, and make sure we don't just hang
	// on illegal access
	*BCM_CPU_CTRL |= 0x0021;
	bcm_pir_cache = 0;
	bcm_ie_kernel = 0;
	ms_im_cache = 0;
	ms_him_cache = 0;
	ms_tim_cache = 0;
	ms_gim_cache = 0;

	// Nasty nasty hack to allow this interrupt to be shared
	// between the kernel cascade function and a device driver.
	// We won't see it in do_IRQ unless something actually
	// turns on mask bits that the kernel didn't:
	tivo_setup_irq(CPU_IRQ_INT2, &cascade_shared);
	tivo_setup_irq(BCM_IRQ_PCI_INT, &cascade);
	tivo_setup_irq(BCM_IRQ_ILL_INT, &bcm_illcyc_irq);
}

void do_7020irq(struct pt_regs *regs)
{
	int i;
	unsigned int intr_stat = *BCM_INTR_STATUS & *BCM_INTR_ENA;

	if (intr_stat & 0x80000000) {
		unsigned int intr_reg = *BCM_PCI_INTR_REG;

		intr_reg &= *BCM_PCI_INTR_STAT;

		*BCM_INTR_CLEAR = 0x80000000;
		for (i = 0; i < 16; i++)
			if (intr_reg & (1 << i)) {
				do_IRQ(i + 40, regs);
				return;
			}
		if (DEBUG_IRQS >= 1) {
			printk(KERN_WARNING "spurious PCI int! BCM_PCI_INTR_REG was 0x%08x is now 0x%08x\n", intr_reg, *BCM_PCI_INTR_REG);
		}
		atomic_inc(&irq_err_count);
		return;
	} else if (intr_stat & bcm_ie_kernel) {
		for (i = 30; i >= 0; i--)
			if (intr_stat & bcm_ie_kernel & (1 << i)) {
				do_IRQ(i + 8, regs);
				return;
			}
	} else if (intr_stat) {
		do_IRQ(CPU_IRQ_INT2, regs);
		return;
	}
	if (DEBUG_IRQS >= 1) {
		printk(KERN_WARNING "spurious BCM int! BCM_INTR_STATUS=0x%08x\n", *BCM_INTR_STATUS);
	}
	atomic_inc(&irq_err_count);
}

asmlinkage void do_vector(int irqmask, struct pt_regs *regs)
{
	int i;

	for (i = 0; i < 8; i++)
		if (irqmask & (1 << (i + 8))) {
			if (i == 2) {
				do_fpgairq(regs);
				return;
			}
			if (i == 4) {
				do_7020irq(regs);
				return;
			}
			do_IRQ(i, regs);
			return;
		}

	if (DEBUG_IRQS >= 4) {
	// This was slowing down the local bus UARTs' interrupt too much,
	// causing serial port FIFO overflow (and thus crypto failure).
	// It needs to be investigated why we're getting spurious ints,
	// but only spew this warning if DEBUG_IRQ for now...
		printk(KERN_WARNING "do_vector got spurious int with mask 0x%08x!\n", irqmask);
	}
	atomic_inc(&irq_err_count);
}

void init_irq_proc (void)
{
}
