/* 
 * This is to write out to the disk in the case that we have
 * essentially a non-functional TiVo.  The idea here is we're
 * going to write out ot the swap partition some log of what
 * happened.  The two cases to get us here generally are
 * a kernel panic, or a watch dog fire.
 *
 * We are going to write out to swap, on boot up, we'll scan
 * the swap before turning it on to see our signature.  If
 * the signature is present, then we'll extract the data
 * and put it in our normal logs.
 *
 * Copyright TiVo (C) 2005
 * Written by Shane Nay (snay@tivo.com)
 *
 */

#include <linux/types.h>
#include <linux/ide.h>
#include <linux/major.h>
#include <linux/delay.h>
#include <linux/ide-deadsystem.h>


#define SWAPFS_MAJOR IDE0_MAJOR
#define SWAPFS_MINOR 8

static int pio_busy_wait(unsigned int usecs, ide_drive_t* drive) {
	int counter = 0;
	while(GET_STAT() & BUSY_STAT) {
		udelay(1);
		if(counter++ > usecs) { 
			printk("IDE Dead System: permanent busy state\n");
			return -ENODEV;
		}
	}
	return 0;
}

static int pio_wait(ide_drive_t* drive) {
	pio_busy_wait(100000, drive); /* 100 ms */
	if(!OK_STAT(GET_STAT(), READY_STAT,
		    BUSY_STAT|WRERR_STAT /* DF */ |DRQ_STAT|ERR_STAT)) {
		printk("IDE Dead System: drive in bad state 0x%x\n", GET_STAT());
		return -ENODEV;
	}
	return 0;
}

void inline lba48_address_write(ide_drive_t* drive, unsigned long sector) {
	OUT_BYTE(0 , IDE_NSECTOR_REG);
	OUT_BYTE(1 , IDE_NSECTOR_REG);
	OUT_BYTE(0 , IDE_HCYL_REG); /* high */
	OUT_BYTE(0 , IDE_LCYL_REG); /* mid */
	OUT_BYTE((sector>>24) & 0xff , IDE_SECTOR_REG);/* low */

	OUT_BYTE((sector>>16) & 0xff , IDE_HCYL_REG);  /* high */
	OUT_BYTE((sector>>8 ) & 0xff , IDE_LCYL_REG);  /* mid */
	OUT_BYTE(sector & 0xff     , IDE_SECTOR_REG);  /* low */
	TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(0x40);
	
	TIVO_OUT_BYTE_IDE_COMMAND_REG_VALUE(0x34);
}

void inline lba28_address_write(ide_drive_t* drive, unsigned long sector) {
	OUT_BYTE(1 , IDE_NSECTOR_REG);
	OUT_BYTE(sector & 0xff       , IDE_SECTOR_REG); /* low */
	OUT_BYTE((sector>>8 ) & 0xff , IDE_LCYL_REG);   /* mid */
	OUT_BYTE((sector>>16) & 0xff , IDE_HCYL_REG);   /* high */ 
	TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(((sector>>24) & 0xf) | 0x40); /* highest */
	
	TIVO_OUT_BYTE_IDE_COMMAND_REG_VALUE(0x30);
}

#if defined (CONFIG_TIVO_EIGER) || defined (CONFIG_TIVO_FUSION)

/* This is so broken, S2 is broken, but we treat everything else like its 
 * broken and S2 is cheery.  Fantabulous.  Hrmph.
 */

#define SF_OUT_WORD(data, reg) OUT_WORD(swab16(data), reg)
#else
#define SF_OUT_WORD(data, reg) OUT_WORD(data, reg)
#endif

int pio_write_buffer(ide_drive_t* drive, unsigned long sector,
		     const u8* buf, unsigned int size, u32 *csum) {
	int ret;
	int writtensects=0;
	int i;
	u32 checksum = 0, carry = 0;
	
	sata_pm_lock_port_exclude_ata_nop(drive);

	while(size > 0) {
		/* Written according to ATA - 6 specification */
		int cnt=0;
		if ((drive->id->cfs_enable_2 & 0x0400)
		    && (drive->addressing)) {
			/* LBA 48 addressing */
			lba48_address_write(drive, sector);
		} else if (drive->select.b.lba) {
			/* LBA 28 addressing */
			lba28_address_write(drive, sector);
		} else {
			sata_pm_unlock_port(drive);
			printk("IDE Dead System doesn't do CHS\n");
			return -ENODEV;
		}

		udelay(1);

		if((ret = pio_busy_wait(100000, drive)) < 0) {
			sata_pm_unlock_port(drive);
			return ret;
		}

		for(i=(512/2);i;i--) { 
			u16 data;
			/* Unaligned access unwise, makes for funky code */
			if(size >= 2) {
				data = *buf++;
				data |= (*buf++)<<8;
				size -= 2;
			} else if (size == 1){
				data = *buf++;
				size = 0;
			} else {
				data = 0;
			}
			
			SF_OUT_WORD(data, IDE_DATA_REG);
			checksum += carry;
			checksum += data;
			carry = (data > checksum);
			cnt++;
		}
		mdelay(5);

		if((ret = pio_wait(drive)) < 0) {
			sata_pm_unlock_port(drive);
			return ret;
		}
;
		sector++;
		writtensects++;
	}
	sata_pm_unlock_port(drive);

	checksum += carry;
	
	if(csum != NULL) {
		*csum += checksum;
		if(checksum > *csum)
			*csum = (*csum) + 1;
	}
	
	return writtensects;
}

static int pio_reset(ide_drive_t* drive) {
	int retries=10;

	sata_pm_lock_port_exclude_ata_nop(drive);
	while(--retries) {
		/* Delays are ATAPI-6 spec times 5 */
		OUT_BYTE(/* nIEN */ 2 | /* SRST */ 4, IDE_CONTROL_REG);
#ifdef CONFIG_TIVO_SATAPM
		sata_pm_ide_intr_pending = 1;
#endif // CONFIG_TIVO_SATAPM
		udelay(50);
		OUT_BYTE(/* nIEN */ 2, IDE_CONTROL_REG);
#ifdef CONFIG_TIVO_SATAPM
		sata_pm_ide_intr_pending = 1;
#endif // CONFIG_TIVO_SATAPM
		mdelay(10);
		if(pio_busy_wait(250000, drive) == 0) { /* Up to 250 ms */
			if(OK_STAT(GET_STAT(), 0,
				    DRQ_STAT|ERR_STAT)) {
				break;
			}
		}
	}

	if(retries == 0) {
		sata_pm_unlock_port(drive);
		printk("IDE Dead System: drive won't reset, giving up\n");
		return -ENODEV;
	}

	TIVO_OUT_BYTE_IDE_SELECT_REG_VALUE(0);
	udelay(1);
	if(pio_wait(drive) < 0) { // Device ready?
		sata_pm_unlock_port(drive);
		return -ENODEV;
	}
	/*
	 * According to the spec, that's all we *have* to do
	 * since we already have the info about the drive that
	 * the kernel queried out.  If this gets buggy on some
	 * drives, we might try adding an IDENTIFY command (0xec
	 * I think)
	 */
	sata_pm_unlock_port(drive);

	return 0;
}

static int initdrive=1;

/* Returns number of sectors written */
int pio_write_buffer_to_swap(const u8* buf, unsigned int size, int secoffset,
	u32 *csum) {
	ide_drive_t* drive= get_info_ptr(MKDEV(SWAPFS_MAJOR, SWAPFS_MINOR));
	struct gendisk *gp= get_gendisk(MKDEV(SWAPFS_MAJOR, 0));
	unsigned long tsector;

	if(gp == NULL || drive == NULL)
		return -ENODEV;

	/*
	 * We adjust secoffset to write to the last 1MB of the device,
	 * this is because we don't want to rebuild swap mostly, and
	 * because we can use "lspanic" to grab the last tivo panic.
	 *
	 * We will track which one we've logged by bit-inverting
	 * the checksum.
	 *
	 */

	/* Partition must be at least 2 MB */
	if(gp->part[SWAPFS_MINOR].nr_sects < (2* 1024 * 1024 / 512))
		return -ENODEV;

	/* Offset by 1MB from end of partition */
	secoffset += gp->part[SWAPFS_MINOR].nr_sects - (1 * 1024 * 1024 /512);

	/* Make sure partition is sane size-wise */

	if(gp->part[SWAPFS_MINOR].nr_sects < secoffset)
		return -ENODEV;
	if(gp->part[SWAPFS_MINOR].nr_sects - secoffset < ((size + 511)/512)) {
		return -ENODEV;
	}
	
	if(initdrive) {
                /* Issue software reset, disable ide interrupts */
		if(pio_reset(drive) < 0) { 
			return -ENODEV;
		}
		initdrive=0;
	}

	tsector = gp->part[SWAPFS_MINOR].start_sect + secoffset;

	return pio_write_buffer(drive, tsector, buf, size, csum);

}


#define CORE_START (0x80000000)
#define CORE_SIZE  (32 * 1024 * 1024)

int pio_write_corefile(int offset) {
	/*
	 * The original idea here is to write to the "alternate
	 * application partition".  This has some problems though,
	 * spec. with developer boxes.  For now, we're going to
	 * assume this won't be called from a dev build.  (This is so
	 * broken it bothers me, we could write it to swap -32M -1M)
	 *
	 * The second broken thing here is we don't account for
	 * actual size of the memory of the box in question.  We
	 * assume it's 32Meg.  This is somewhat easily overcome by
	 * just checking the config variables.  However, I'm loathe
	 * to do that at this point since having variable size input
	 * just risks adding a new bug to the system.  (Something I don't
	 * want to do at this point in the release cycle)
	 *
	 * This is a *very basic* implementation of what we want to do.
	 * It starts at 0x80000000, goes for 32MB writing out to alternate
	 * application at either "offset 0", or "offset 1".  Offset 0 is
	 * at the beginning of the partition.  Offset 1 is 32MB into the
	 * partition.  (ah, now you see why I don't want to deal with
	 * variable sizes)
	 *
	 * Cache flushing is not required as we're doing it PIO, which just
	 * pulls from cache.  Of course, we're taking the kernel centric
	 * as opposed to hardware centric view of memory.  This is reasonable
	 * because really, the core is just a bunch of kernel memory
	 * structures for the most part.
	 * 
	 */
	kdev_t altdev;
	int altpartition;
	ide_drive_t* drive;
	struct gendisk *gp= get_gendisk(MKDEV(IDE0_MAJOR, 0));
	unsigned long secoffset = (CORE_SIZE / 512) * offset;
	u32 csum; /* Currently unused */
	static int core_written = 0;

	if(core_written != 0) {
		printk("Core already written\n");
		return -EINVAL;
	}
	
	core_written = 1;
	
	/* Only 0 or 1 are valid offset values */
	if(!(offset == 0 || offset == 1))
		return -EINVAL;


	/* If the root is hda4, then altap is hda7 and vice-versa */
	if(ROOT_DEV == MKDEV(IDE0_MAJOR, 4))
		altpartition = 7;
	else if(ROOT_DEV == MKDEV(IDE0_MAJOR, 7))
		altpartition = 4;
	else
		return -ENODEV;

	altdev = MKDEV(IDE0_MAJOR, altpartition);

	drive = get_info_ptr(altdev);

	if(gp == NULL || drive == NULL)
		return -ENODEV;

	/* Partition must be at least twice a CORE */
	if(gp->part[altpartition].nr_sects < ((2 * CORE_SIZE) / 512))
		return -ENODEV;
	
	if(initdrive) {
                /* Issue software reset, disable ide interrupts */
		if(pio_reset(drive) < 0) { 
			return -ENODEV;
		}
		initdrive=0;
	}

	secoffset += gp->part[altpartition].start_sect;
	
	return pio_write_buffer(drive, secoffset, (const u8*)CORE_START,
		       CORE_SIZE, &csum);
}
